@@ -323,6 +323,15 @@ def lock(ctx, state, **kwargs):
     dev_only = state.lockoptions.dev_only
     pre = state.installstate.pre
     if emit_requirements:
+        secho(
+            "Warning: The lock flag -r/--requirements will be deprecated in a future version\n"
+            "of pipenv in favor of the new requirements command. For more info see\n"
+            "https://pipenv.pypa.io/en/latest/advanced/#generating-a-requirements-txt\n"
+            "NOTE: the requirements command parses Pipfile.lock directly without performing any\n"
+            "locking operations. Updating packages should be done by running pipenv lock.",
+            fg="yellow",
+            err=True,
+        )
         # Emit requirements file header (unless turned off with --no-header)
         if state.lockoptions.emit_requirements_header:
             header_options = ["--requirements"]
@@ -754,26 +763,37 @@ def verify(state):
     "--dev-only", is_flag=True, default=False, help="Only add development requirements."
 )
 @option("--hash", is_flag=True, default=False, help="Add package hashes.")
+@option("--exclude-markers", is_flag=True, default=False, help="Exclude markers.")
 @pass_state
-def requirements(state, dev=False, dev_only=False, hash=False):
-    lockfile = state.project.lockfile_content
+def requirements(state, dev=False, dev_only=False, hash=False, exclude_markers=False):
+
+    from pipenv.utils.dependencies import convert_deps_to_pip
+
+    lockfile = state.project.load_lockfile(expand_env_vars=False)
+
     for i, package_index in enumerate(lockfile["_meta"]["sources"]):
         prefix = "-i" if i == 0 else "--extra-index-url"
-        echo(crayons.normal(" ".join([prefix, package_index["url"]])))
-    if not dev_only:
-        for req_name, value in lockfile["default"].items():
-            if hash:
-                hashes = [f" \\\n    --hash={h}" for h in value.get("hashes", [])]
-            else:
-                hashes = []
-            echo(crayons.normal("".join([req_name, value["version"], *hashes])))
+        echo(" ".join([prefix, package_index["url"]]))
+
+    deps = {}
+
     if dev or dev_only:
-        for req_name, value in lockfile["develop"].items():
-            if hash:
-                hashes = [f" \\\n    --hash={h}" for h in value.get("hashes", [])]
-            else:
-                hashes = []
-            echo(crayons.normal("".join([req_name, value["version"], *hashes])))
+        deps.update(lockfile["develop"])
+    if not dev_only:
+        deps.update(lockfile["default"])
+
+    pip_deps = convert_deps_to_pip(
+        deps,
+        project=None,
+        r=False,
+        include_index=False,
+        include_hashes=hash,
+        include_markers=not exclude_markers,
+    )
+
+    for d in pip_deps:
+        echo(d)
+
     sys.exit(0)
 
 
