@@ -1,7 +1,8 @@
 # -*- coding: utf-8 -*-
+from __future__ import annotations
+
 import base64
 import fnmatch
-import glob
 import hashlib
 import io
 import json
@@ -9,36 +10,50 @@ import operator
 import os
 import re
 import sys
+import urllib.parse
+from pathlib import Path
 
-import six
+import pipfile
+import pipfile.api
 import toml
 import tomlkit
 import vistir
 
-import pipfile
-import pipfile.api
-
-from .vendor.cached_property import cached_property
-
-from .cmdparse import Script
-from .environment import Environment
-from .environments import (
-    PIPENV_DEFAULT_PYTHON_VERSION, PIPENV_IGNORE_VIRTUALENVS, PIPENV_MAX_DEPTH,
-    PIPENV_PIPFILE, PIPENV_PYTHON, PIPENV_TEST_INDEX, PIPENV_VENV_IN_PROJECT,
-    PIPENV_USE_SYSTEM, is_in_virtualenv, is_type_checking
+from pipenv.cmdparse import Script
+from pipenv.core import system_which
+from pipenv.environment import Environment
+from pipenv.environments import (
+    Setting,
+    is_in_virtualenv,
+    is_type_checking,
+    normalize_pipfile_path,
 )
-from .vendor.requirementslib.models.utils import get_default_pyproject_backend
-from .utils import (
-    cleanup_toml, convert_toml_outline_tables, find_requirements,
-    get_canonical_names, get_url_name, get_workon_home, is_editable,
-    is_installable_file, is_star, is_valid_url, is_virtual_environment,
-    looks_like_dir, normalize_drive, pep423_name, proper_case, python_version,
-    safe_expandvars, get_pipenv_dist
+from pipenv.utils.dependencies import (
+    get_canonical_names,
+    is_editable,
+    is_star,
+    python_version,
 )
+from pipenv.utils.internet import get_url_name, is_valid_url, proper_case
+from pipenv.utils.resolver import pep423_name
+from pipenv.utils.shell import (
+    find_requirements,
+    find_windows_executable,
+    get_pipenv_dist,
+    get_workon_home,
+    is_virtual_environment,
+    looks_like_dir,
+    safe_expandvars,
+)
+from pipenv.utils.toml import cleanup_toml, convert_toml_outline_tables
+from pipenv.vendor.cached_property import cached_property
+from pipenv.vendor.requirementslib.models.utils import get_default_pyproject_backend
 
 if is_type_checking():
-    import pkg_resources
     from typing import Dict, List, Optional, Set, Text, Tuple, Union
+
+    import pkg_resources
+
     TSource = Dict[Text, Union[Text, bool]]
     TPackageEntry = Dict[str, Union[bool, str, List[str]]]
     TPackage = Dict[str, TPackageEntry]
@@ -47,24 +62,7 @@ if is_type_checking():
     TPipfile = Dict[str, Union[TPackage, TScripts, TPipenv, List[TSource]]]
 
 
-def _normalized(p):
-    if p is None:
-        return None
-    loc = vistir.compat.Path(p)
-    try:
-        loc = loc.resolve()
-    except OSError:
-        loc = loc.absolute()
-    # Recase the path properly on Windows. From https://stackoverflow.com/a/35229734/5043728
-    if os.name == 'nt':
-        matches = glob.glob(re.sub(r'([^:/\\])(?=[/\\]|$)', r'[\1]', str(loc)))
-        path_str = matches and matches[0] or str(loc)
-    else:
-        path_str = str(loc)
-    return normalize_drive(os.path.abspath(path_str))
-
-
-DEFAULT_NEWLINES = u"\n"
+DEFAULT_NEWLINES = "\n"
 
 
 class _LockFileEncoder(json.JSONEncoder):
@@ -83,64 +81,38 @@ class _LockFileEncoder(json.JSONEncoder):
         )
 
     def default(self, obj):
-        if isinstance(obj, vistir.compat.Path):
+        if isinstance(obj, Path):
             obj = obj.as_posix()
         return super(_LockFileEncoder, self).default(obj)
 
     def encode(self, obj):
         content = super(_LockFileEncoder, self).encode(obj)
-        if not isinstance(content, six.text_type):
+        if not isinstance(content, str):
             content = content.decode("utf-8")
         return content
 
 
 def preferred_newlines(f):
-    if isinstance(f.newlines, six.text_type):
+    if isinstance(f.newlines, str):
         return f.newlines
     return DEFAULT_NEWLINES
 
 
-if PIPENV_PIPFILE:
-    if not os.path.isfile(PIPENV_PIPFILE):
-        raise RuntimeError("Given PIPENV_PIPFILE is not found!")
-
-    else:
-        PIPENV_PIPFILE = _normalized(PIPENV_PIPFILE)
-        # Overwrite environment variable so that subprocesses can get the correct path.
-        # See https://github.com/pypa/pipenv/issues/3584
-        os.environ['PIPENV_PIPFILE'] = PIPENV_PIPFILE
 # (path, file contents) => TOMLFile
 # keeps track of pipfiles that we've seen so we do not need to re-parse 'em
 _pipfile_cache = {}
 
 
-if PIPENV_TEST_INDEX:
-    DEFAULT_SOURCE = {
-        u"url": PIPENV_TEST_INDEX,
-        u"verify_ssl": True,
-        u"name": u"custom",
-    }
-else:
-    DEFAULT_SOURCE = {
-        u"url": u"https://pypi.org/simple",
-        u"verify_ssl": True,
-        u"name": u"pypi",
-    }
-
-pipfile.api.DEFAULT_SOURCE = DEFAULT_SOURCE
-
-
 class SourceNotFound(KeyError):
     pass
 
 
-class Project(object):
+class Project:
     """docstring for Project"""
 
     _lockfile_encoder = _LockFileEncoder()
 
-    def __init__(self, which=None, python_version=None, chdir=True):
-        super(Project, self).__init__()
+    def __init__(self, python_version=None, chdir=True):
         self._name = None
         self._virtualenv_location = None
         self._download_location = None
@@ -151,11 +123,23 @@ class Project(object):
         self._requirements_location = None
         self._original_dir = os.path.abspath(os.curdir)
         self._environment = None
-        self._which = which
-        self._build_system = {
-            "requires": ["setuptools", "wheel"]
-        }
+        self._build_system = {"requires": ["setuptools", "wheel"]}
         self.python_version = python_version
+        self.s = Setting()
+        if self.s.PIPENV_TEST_INDEX:
+            self.default_source = {
+                "url": self.s.PIPENV_TEST_INDEX,
+                "verify_ssl": True,
+                "name": "custom",
+            }
+        else:
+            self.default_source = {
+                "url": "https://pypi.org/simple",
+                "verify_ssl": True,
+                "name": "pypi",
+            }
+        pipfile.api.DEFAULT_SOURCE = self.default_source
+
         # Hack to skip this during pipenv run, or -r.
         if ("run" not in sys.argv) and chdir:
             try:
@@ -163,8 +147,7 @@ class Project(object):
             except (TypeError, AttributeError):
                 pass
 
-    def path_to(self, p):
-        # type: (str) -> str
+    def path_to(self, p: str) -> str:
         """Returns the absolute path to a given relative path."""
         if os.path.isabs(p):
             return p
@@ -172,102 +155,46 @@ class Project(object):
         return os.sep.join([self._original_dir, p])
 
     def _build_package_list(self, package_section):
-        """Returns a list of packages for pip-tools to consume."""
-        from pipenv.vendor.requirementslib.utils import is_vcs
-        ps = {}
-        # TODO: Separate the logic for showing packages from the filters for supplying pip-tools
-        for k, v in self.parsed_pipfile.get(package_section, {}).items():
-            # Skip editable VCS deps.
-            if hasattr(v, "keys"):
-                # When a vcs url is gven without editable it only appears as a key
-                # Eliminate any vcs, path, or url entries which are not editable
-                # Since pip-tools can't do deep resolution on them, even setuptools-installable ones
-                if (
-                    is_vcs(v)
-                    or is_vcs(k)
-                    or (is_installable_file(k) or is_installable_file(v))
-                    or any(
-                        (
-                            prefix in v
-                            and (os.path.isfile(v[prefix]) or is_valid_url(v[prefix]))
-                        )
-                        for prefix in ["path", "file"]
-                    )
-                ):
-                    # If they are editable, do resolve them
-                    if "editable" not in v:
-                        # allow wheels to be passed through
-                        if not (
-                            hasattr(v, "keys")
-                            and v.get("path", v.get("file", "")).endswith(".whl")
-                        ):
-                            continue
-                        ps.update({k: v})
-
-                    else:
-                        ps.update({k: v})
-                else:
-                    ps.update({k: v})
-            else:
-                # Since these entries have no attributes we know they are not editable
-                # So we can safely exclude things that need to be editable in order to be resolved
-                # First exclude anything that is a vcs entry either in the key or value
-                if not (
-                    any(is_vcs(i) for i in [k, v])
-                    # Then exclude any installable files that are not directories
-                    # Because pip-tools can resolve setup.py for example
-                    or any(is_installable_file(i) for i in [k, v])
-                    # Then exclude any URLs because they need to be editable also
-                    # Things that are excluded can only be 'shallow resolved'
-                    or any(is_valid_url(i) for i in [k, v])
-                ):
-                    ps.update({k: v})
-        return ps
+        """Returns the list of packages from the Project's Pipfile."""
+        return self.parsed_pipfile.get(package_section, {})
 
     @property
-    def name(self):
-        # type: () -> str
+    def name(self) -> str:
         if self._name is None:
             self._name = self.pipfile_location.split(os.sep)[-2]
         return self._name
 
     @property
-    def pipfile_exists(self):
-        # type: () -> bool
+    def pipfile_exists(self) -> bool:
         return os.path.isfile(self.pipfile_location)
 
     @property
-    def required_python_version(self):
-        # type: () -> str
+    def required_python_version(self) -> str:
         if self.pipfile_exists:
-            required = self.parsed_pipfile.get("requires", {}).get(
-                "python_full_version"
-            )
+            required = self.parsed_pipfile.get("requires", {}).get("python_full_version")
             if not required:
                 required = self.parsed_pipfile.get("requires", {}).get("python_version")
             if required != "*":
                 return required
 
     @property
-    def project_directory(self):
-        # type: () -> str
+    def project_directory(self) -> str:
         return os.path.abspath(os.path.join(self.pipfile_location, os.pardir))
 
     @property
-    def requirements_exists(self):
-        # type: () -> bool
+    def requirements_exists(self) -> bool:
         return bool(self.requirements_location)
 
-    def is_venv_in_project(self):
-        # type: () -> bool
-        return PIPENV_VENV_IN_PROJECT or (
+    def is_venv_in_project(self) -> bool:
+        if self.s.PIPENV_VENV_IN_PROJECT is False:
+            return False
+        return self.s.PIPENV_VENV_IN_PROJECT or (
             self.project_directory
             and os.path.isdir(os.path.join(self.project_directory, ".venv"))
         )
 
     @property
-    def virtualenv_exists(self):
-        # type: () -> bool
+    def virtualenv_exists(self) -> bool:
         if os.path.exists(self.virtualenv_location):
             if os.name == "nt":
                 extra = ["Scripts", "activate.bat"]
@@ -277,8 +204,7 @@ class Project(object):
 
         return False
 
-    def get_location_for_virtualenv(self):
-        # type: () -> str
+    def get_location_for_virtualenv(self) -> str:
         # If there's no project yet, set location based on config.
         if not self.project_directory:
             if self.is_venv_in_project():
@@ -301,19 +227,24 @@ class Project(object):
         with io.open(dot_venv) as f:
             name = f.read().strip()
 
+        # If .venv file is empty, set location based on config.
+        if not name:
+            return str(get_workon_home().joinpath(self.virtualenv_name))
+
         # If content looks like a path, use it as a relative path.
-        # Otherwise use directory named after content in WORKON_HOME.
+        # Otherwise, use directory named after content in WORKON_HOME.
         if looks_like_dir(name):
-            path = vistir.compat.Path(self.project_directory, name)
+            path = Path(self.project_directory, name)
             return path.absolute().as_posix()
         return str(get_workon_home().joinpath(name))
 
     @property
-    def working_set(self):
-        # type: () -> pkg_resources.WorkingSet
-        from .utils import load_path
+    def working_set(self) -> pkg_resources.WorkingSet:
+        from pipenv.utils.shell import load_path
+
         sys_path = load_path(self.which("python"))
         import pkg_resources
+
         return pkg_resources.WorkingSet(sys_path)
 
     @property
@@ -321,43 +252,45 @@ class Project(object):
         return self.environment.get_installed_packages()
 
     @property
-    def installed_package_names(self):
-        # type: () -> List[str]
+    def installed_package_names(self) -> List[str]:
         return get_canonical_names([pkg.key for pkg in self.installed_packages])
 
     @property
-    def lockfile_package_names(self):
-        # type: () -> Dict[str, Set[str]]
+    def lockfile_package_names(self) -> Dict[str, Set[str]]:
         dev_keys = get_canonical_names(self.lockfile_content["develop"].keys())
         default_keys = get_canonical_names(self.lockfile_content["default"].keys())
         return {
             "dev": dev_keys,
             "default": default_keys,
-            "combined": dev_keys | default_keys
+            "combined": dev_keys | default_keys,
         }
 
     @property
-    def pipfile_package_names(self):
-        # type: () -> Dict[str, Set[str]]
+    def pipfile_package_names(self) -> Dict[str, Set[str]]:
         dev_keys = get_canonical_names(self.dev_packages.keys())
         default_keys = get_canonical_names(self.packages.keys())
         return {
             "dev": dev_keys,
             "default": default_keys,
-            "combined": dev_keys | default_keys
+            "combined": dev_keys | default_keys,
         }
 
-    def get_environment(self, allow_global=False):
-        # type: (bool) -> Environment
+    def get_environment(self, allow_global: bool = False) -> Environment:
         is_venv = is_in_virtualenv()
         if allow_global and not is_venv:
             prefix = sys.prefix
+            python = sys.executable
         else:
             prefix = self.virtualenv_location
-        sources = self.sources if self.sources else [DEFAULT_SOURCE]
+            python = None
+        sources = self.sources if self.sources else [self.default_source]
         environment = Environment(
-            prefix=prefix, is_venv=is_venv, sources=sources, pipfile=self.parsed_pipfile,
-            project=self
+            prefix=prefix,
+            python=python,
+            is_venv=is_venv,
+            sources=sources,
+            pipfile=self.parsed_pipfile,
+            project=self,
         )
         pipenv_dist = get_pipenv_dist(pkg="pipenv")
         if pipenv_dist:
@@ -367,20 +300,17 @@ class Project(object):
         return environment
 
     @property
-    def environment(self):
-        # type: () -> Environment
+    def environment(self) -> Environment:
         if not self._environment:
-            allow_global = os.environ.get("PIPENV_USE_SYSTEM", PIPENV_USE_SYSTEM)
+            allow_global = self.s.PIPENV_USE_SYSTEM
             self._environment = self.get_environment(allow_global=allow_global)
         return self._environment
 
-    def get_outdated_packages(self):
-        # type: () -> List[pkg_resources.Distribution]
+    def get_outdated_packages(self) -> List[pkg_resources.Distribution]:
         return self.environment.get_outdated_packages(pre=self.pipfile.get("pre", False))
 
     @classmethod
-    def _sanitize(cls, name):
-        # type: (str) -> Tuple[str, str]
+    def _sanitize(cls, name: str) -> Tuple[str, str]:
         # Replace dangerous characters into '_'. The length of the sanitized
         # project name is limited as 42 because of the limit of linux kernel
         #
@@ -393,10 +323,9 @@ class Project(object):
         #   https://www.gnu.org/software/bash/manual/html_node/Double-Quotes.html
         #   http://www.tldp.org/LDP/abs/html/special-chars.html#FIELDREF
         #   https://github.com/torvalds/linux/blob/2bfe01ef/include/uapi/linux/binfmts.h#L18
-        return re.sub(r'[ $`!*@"\\\r\n\t]', "_", name)[0:42]
+        return re.sub(r'[ &$`!*@"()\[\]\\\r\n\t]', "_", name)[0:42]
 
-    def _get_virtualenv_hash(self, name):
-        # type: (str) -> str
+    def _get_virtualenv_hash(self, name: str) -> str:
         """Get the name of the virtualenv adjusted for windows if needed
 
         Returns (name, encoded_hash)
@@ -438,22 +367,27 @@ class Project(object):
         return clean_name, encoded_hash
 
     @property
-    def virtualenv_name(self):
-        # type: () -> str
+    def virtualenv_name(self) -> str:
         sanitized, encoded_hash = self._get_virtualenv_hash(self.name)
-        suffix = "-{0}".format(PIPENV_PYTHON) if PIPENV_PYTHON else ""
+        suffix = ""
+        if self.s.PIPENV_PYTHON:
+            if os.path.isabs(self.s.PIPENV_PYTHON):
+                suffix = "-{0}".format(os.path.basename(self.s.PIPENV_PYTHON))
+            else:
+                suffix = "-{0}".format(self.s.PIPENV_PYTHON)
+
         # If the pipfile was located at '/home/user/MY_PROJECT/Pipfile',
         # the name of its virtualenv will be 'my-project-wyUfYPqE'
         return sanitized + "-" + encoded_hash + suffix
 
     @property
-    def virtualenv_location(self):
-        # type: () -> str
+    def virtualenv_location(self) -> str:
         # if VIRTUAL_ENV is set, use that.
         virtualenv_env = os.getenv("VIRTUAL_ENV")
         if (
             "PIPENV_ACTIVE" not in os.environ
-            and not PIPENV_IGNORE_VIRTUALENVS and virtualenv_env
+            and not self.s.PIPENV_IGNORE_VIRTUALENVS
+            and virtualenv_env
         ):
             return virtualenv_env
 
@@ -463,8 +397,7 @@ class Project(object):
         return self._virtualenv_location
 
     @property
-    def virtualenv_src_location(self):
-        # type: () -> str
+    def virtualenv_src_location(self) -> str:
         if self.virtualenv_location:
             loc = os.sep.join([self.virtualenv_location, "src"])
         else:
@@ -473,8 +406,7 @@ class Project(object):
         return loc
 
     @property
-    def download_location(self):
-        # type: () -> str
+    def download_location(self) -> str:
         if self._download_location is None:
             loc = os.sep.join([self.virtualenv_location, "downloads"])
             self._download_location = loc
@@ -483,55 +415,49 @@ class Project(object):
         return self._download_location
 
     @property
-    def proper_names_db_path(self):
-        # type: () -> str
+    def proper_names_db_path(self) -> str:
         if self._proper_names_db_path is None:
-            self._proper_names_db_path = vistir.compat.Path(
+            self._proper_names_db_path = Path(
                 self.virtualenv_location, "pipenv-proper-names.txt"
             )
         self._proper_names_db_path.touch()  # Ensure the file exists.
         return self._proper_names_db_path
 
     @property
-    def proper_names(self):
-        # type: () -> str
+    def proper_names(self) -> str:
         with self.proper_names_db_path.open() as f:
             return f.read().splitlines()
 
-    def register_proper_name(self, name):
-        # type: (str) -> None
+    def register_proper_name(self, name: str) -> None:
         """Registers a proper name to the database."""
         with self.proper_names_db_path.open("a") as f:
-            f.write(u"{0}\n".format(name))
+            f.write("{0}\n".format(name))
 
     @property
-    def pipfile_location(self):
-        # type: () -> str
-        if PIPENV_PIPFILE:
-            return PIPENV_PIPFILE
+    def pipfile_location(self) -> str:
+        if self.s.PIPENV_PIPFILE:
+            return self.s.PIPENV_PIPFILE
 
         if self._pipfile_location is None:
             try:
-                loc = pipfile.Pipfile.find(max_depth=PIPENV_MAX_DEPTH)
+                loc = pipfile.Pipfile.find(max_depth=self.s.PIPENV_MAX_DEPTH)
             except RuntimeError:
                 loc = "Pipfile"
-            self._pipfile_location = _normalized(loc)
+            self._pipfile_location = normalize_pipfile_path(loc)
         return self._pipfile_location
 
     @property
-    def requirements_location(self):
-        # type: () -> Optional[str]
+    def requirements_location(self) -> Optional[str]:
         if self._requirements_location is None:
             try:
-                loc = find_requirements(max_depth=PIPENV_MAX_DEPTH)
+                loc = find_requirements(max_depth=self.s.PIPENV_MAX_DEPTH)
             except RuntimeError:
                 loc = None
             self._requirements_location = loc
         return self._requirements_location
 
     @property
-    def parsed_pipfile(self):
-        # type: () -> Union[tomlkit.toml_document.TOMLDocument, TPipfile]
+    def parsed_pipfile(self) -> Union[tomlkit.toml_document.TOMLDocument, TPipfile]:
         """Parse Pipfile into a TOMLFile and cache it
 
         (call clear_pipfile_cache() afterwards if mutating)"""
@@ -543,8 +469,7 @@ class Project(object):
             _pipfile_cache[cache_key] = parsed
         return _pipfile_cache[cache_key]
 
-    def read_pipfile(self):
-        # type: () -> str
+    def read_pipfile(self) -> str:
         # Open the pipfile, read it into memory.
         if not self.pipfile_exists:
             return ""
@@ -554,13 +479,13 @@ class Project(object):
 
         return contents
 
-    def clear_pipfile_cache(self):
-        # type: () -> None
+    def clear_pipfile_cache(self) -> None:
         """Clear pipfile cache (e.g., so we can mutate parsed pipfile)"""
         _pipfile_cache.clear()
 
-    def _parse_pipfile(self, contents):
-        # type: () -> Union[tomlkit.toml_document.TOMLDocument, TPipfile]
+    def _parse_pipfile(
+        self, contents: str
+    ) -> Union[tomlkit.toml_document.TOMLDocument, TPipfile]:
         try:
             return tomlkit.parse(contents)
         except Exception:
@@ -568,8 +493,7 @@ class Project(object):
             # Fallback to toml parser, for large files.
             return toml.loads(contents)
 
-    def _read_pyproject(self):
-        # type: () -> None
+    def _read_pyproject(self) -> None:
         pyproject = self.path_to("pyproject.toml")
         if os.path.exists(pyproject):
             self._pyproject = toml.load(pyproject)
@@ -583,30 +507,25 @@ class Project(object):
                 self._build_system = build_system
 
     @property
-    def build_requires(self):
-        # type: () -> List[str]
+    def build_requires(self) -> List[str]:
         return self._build_system.get("requires", ["setuptools>=40.8.0", "wheel"])
 
     @property
-    def build_backend(self):
-        # type: () -> str
+    def build_backend(self) -> str:
         return self._build_system.get("build-backend", get_default_pyproject_backend())
 
     @property
-    def settings(self):
-        # type: () -> Union[tomlkit.items.Table, Dict[str, Union[str, bool]]]
+    def settings(self) -> Union[tomlkit.items.Table, Dict[str, Union[str, bool]]]:
         """A dictionary of the settings added to the Pipfile."""
         return self.parsed_pipfile.get("pipenv", {})
 
-    def has_script(self, name):
-        # type: (str) -> bool
+    def has_script(self, name: str) -> bool:
         try:
             return name in self.parsed_pipfile["scripts"]
         except KeyError:
             return False
 
-    def build_script(self, name, extra_args=None):
-        # type: (str, Optional[List[str]])
+    def build_script(self, name: str, extra_args: Optional[List[str]] = None) -> Script:
         try:
             script = Script.parse(self.parsed_pipfile["scripts"][name])
         except KeyError:
@@ -615,8 +534,7 @@ class Project(object):
             script.extend(extra_args)
         return script
 
-    def update_settings(self, d):
-        # type: (Dict[str, Union[str, bool]]) -> None
+    def update_settings(self, d: Dict[str, Union[str, bool]]) -> None:
         settings = self.settings
         changed = False
         for new in d:
@@ -644,6 +562,7 @@ class Project(object):
     @property
     def _pipfile(self):
         from .vendor.requirementslib.models.pipfile import Pipfile as ReqLibPipfile
+
         pf = ReqLibPipfile.load(self.pipfile_location)
         return pf
 
@@ -670,6 +589,7 @@ class Project(object):
 
     def _get_vcs_packages(self, dev=False):
         from pipenv.vendor.requirementslib.utils import is_vcs
+
         section = "dev-packages" if dev else "packages"
         packages = {
             k: v
@@ -688,12 +608,12 @@ class Project(object):
 
     @property
     def vcs_packages(self):
-        """Returns a list of VCS packages, for not pip-tools to consume."""
+        """Returns a list of VCS packages."""
         return self._get_vcs_packages(dev=False)
 
     @property
     def vcs_dev_packages(self):
-        """Returns a list of VCS packages, for not pip-tools to consume."""
+        """Returns a list of VCS packages."""
         return self._get_vcs_packages(dev=True)
 
     @property
@@ -705,12 +625,12 @@ class Project(object):
 
     @property
     def packages(self):
-        """Returns a list of packages, for pip-tools to consume."""
+        """Returns a list of packages."""
         return self._build_package_list("packages")
 
     @property
     def dev_packages(self):
-        """Returns a list of dev-packages, for pip-tools to consume."""
+        """Returns a list of dev-packages."""
         return self._build_package_list("dev-packages")
 
     @property
@@ -718,7 +638,7 @@ class Project(object):
         if not self.pipfile_exists:
             return True
 
-        if not len(self.read_pipfile()):
+        if not self.read_pipfile():
             return True
 
         return False
@@ -726,25 +646,24 @@ class Project(object):
     def create_pipfile(self, python=None):
         """Creates the Pipfile, filled with juicy defaults."""
         from .vendor.pip_shims.shims import InstallCommand
+
         # Inherit the pip's index configuration of install command.
         command = InstallCommand()
         indexes = command.cmd_opts.get_option("--extra-index-url").default
-        sources = [DEFAULT_SOURCE]
+        sources = [self.default_source]
         for i, index in enumerate(indexes):
             if not index:
                 continue
 
             source_name = "pip_index_{}".format(i)
             verify_ssl = index.startswith("https")
-            sources.append(
-                {u"url": index, u"verify_ssl": verify_ssl, u"name": source_name}
-            )
+            sources.append({"url": index, "verify_ssl": verify_ssl, "name": source_name})
 
         data = {
-            u"source": sources,
+            "source": sources,
             # Default packages.
-            u"packages": {},
-            u"dev-packages": {},
+            "packages": {},
+            "dev-packages": {},
         }
         # Default requires.
         required_python = python
@@ -753,9 +672,9 @@ class Project(object):
                 required_python = self.which("python", self.virtualenv_location)
             else:
                 required_python = self.which("python")
-        version = python_version(required_python) or PIPENV_DEFAULT_PYTHON_VERSION
-        if version and len(version) >= 3:
-            data[u"requires"] = {"python_version": version[: len("2.7")]}
+        version = python_version(required_python) or self.s.PIPENV_DEFAULT_PYTHON_VERSION
+        if version and len(version.split(".")) > 2:
+            data["requires"] = {"python_version": ".".join(version.split(".")[:2])}
         self.write_toml(data)
 
     @classmethod
@@ -771,12 +690,15 @@ class Project(object):
         return source
 
     def get_or_create_lockfile(self, from_pipfile=False):
-        from pipenv.vendor.requirementslib.models.lockfile import Lockfile as Req_Lockfile
+        from pipenv.vendor.requirementslib.models.lockfile import (
+            Lockfile as Req_Lockfile,
+        )
+
         lockfile = None
         if from_pipfile and self.pipfile_exists:
             lockfile_dict = {
                 "default": self._lockfile["default"].copy(),
-                "develop": self._lockfile["develop"].copy()
+                "develop": self._lockfile["develop"].copy(),
             }
             lockfile_dict.update({"_meta": self.get_lockfile_meta()})
             lockfile = Req_Lockfile.from_data(
@@ -786,9 +708,13 @@ class Project(object):
             try:
                 lockfile = Req_Lockfile.load(self.lockfile_location)
             except OSError:
-                lockfile = Req_Lockfile.from_data(self.lockfile_location, self.lockfile_content)
+                lockfile = Req_Lockfile.from_data(
+                    self.lockfile_location, self.lockfile_content
+                )
         else:
-            lockfile = Req_Lockfile.from_data(path=self.lockfile_location, data=self._lockfile, meta_from_project=False)
+            lockfile = Req_Lockfile.from_data(
+                path=self.lockfile_location, data=self._lockfile, meta_from_project=False
+            )
         if lockfile._lockfile is not None:
             return lockfile
         if self.lockfile_exists and self.lockfile_content:
@@ -798,9 +724,7 @@ class Project(object):
                 sources = self.pipfile_sources
             elif not isinstance(sources, list):
                 sources = [sources]
-            lockfile_dict["_meta"]["sources"] = [
-                self.populate_source(s) for s in sources
-            ]
+            lockfile_dict["_meta"]["sources"] = [self.populate_source(s) for s in sources]
             _created_lockfile = Req_Lockfile.from_data(
                 path=self.lockfile_location, data=lockfile_dict, meta_from_project=False
             )
@@ -811,17 +735,20 @@ class Project(object):
 
     def get_lockfile_meta(self):
         from .vendor.plette.lockfiles import PIPFILE_SPEC_CURRENT
+
         if self.lockfile_exists:
             sources = self.lockfile_content.get("_meta", {}).get("sources", [])
-        else:
+        elif "source" in self.parsed_pipfile:
             sources = [dict(source) for source in self.parsed_pipfile["source"]]
+        else:
+            sources = self.pipfile_sources
         if not isinstance(sources, list):
             sources = [sources]
         return {
             "hash": {"sha256": self.calculate_pipfile_hash()},
             "pipfile-spec": PIPFILE_SPEC_CURRENT,
             "sources": [self.populate_source(s) for s in sources],
-            "requires": self.parsed_pipfile.get("requires", {})
+            "requires": self.parsed_pipfile.get("requires", {}),
         }
 
     def write_toml(self, data, path=None):
@@ -842,13 +769,12 @@ class Project(object):
                         table.update(data[section][package])
                         document[section][package] = table
                     else:
-                        document[section][package] = tomlkit.string(data[section][package])
+                        document[section][package] = tomlkit.string(
+                            data[section][package]
+                        )
             formatted_data = tomlkit.dumps(document).rstrip()
 
-        if (
-            vistir.compat.Path(path).absolute()
-            == vistir.compat.Path(self.pipfile_location).absolute()
-        ):
+        if Path(path).absolute() == Path(self.pipfile_location).absolute():
             newlines = self._pipfile_newlines
         else:
             newlines = DEFAULT_NEWLINES
@@ -859,8 +785,7 @@ class Project(object):
         self.clear_pipfile_cache()
 
     def write_lockfile(self, content):
-        """Write out the lockfile.
-        """
+        """Write out the lockfile."""
         s = self._lockfile_encoder.encode(content)
         open_kwargs = {"newline": self._lockfile_newlines, "encoding": "utf-8"}
         with vistir.contextmanagers.atomic_open_for_write(
@@ -869,13 +794,13 @@ class Project(object):
             f.write(s)
             # Write newline at end of document. GH-319.
             # Only need '\n' here; the file object handles the rest.
-            if not s.endswith(u"\n"):
-                f.write(u"\n")
+            if not s.endswith("\n"):
+                f.write("\n")
 
     @property
     def pipfile_sources(self):
         if self.pipfile_is_empty or "source" not in self.parsed_pipfile:
-            return [DEFAULT_SOURCE]
+            return [self.default_source]
         # We need to make copies of the source info so we don't
         # accidentally modify the cache. See #2100 where values are
         # written after the os.path.expandvars() call.
@@ -915,19 +840,23 @@ class Project(object):
         return source
 
     def get_source(self, name=None, url=None, refresh=False):
-        from .utils import is_url_equal
+        from pipenv.utils.internet import is_url_equal
 
         def find_source(sources, name=None, url=None):
             source = None
             if name:
-                source = next(iter(
-                    s for s in sources if "name" in s and s["name"] == name
-                ), None)
+                source = next(
+                    iter(s for s in sources if "name" in s and s["name"] == name), None
+                )
             elif url:
-                source = next(iter(
-                    s for s in sources
-                    if "url" in s and is_url_equal(url, s.get("url", ""))
-                ), None)
+                source = next(
+                    iter(
+                        s
+                        for s in sources
+                        if "url" in s and is_url_equal(url, s.get("url", ""))
+                    ),
+                    None,
+                )
             if source is not None:
                 return source
 
@@ -967,9 +896,9 @@ class Project(object):
         packages = set([pep423_name(pkg) for pkg in packages])
         for section in ("dev-packages", "packages"):
             pipfile_section = parsed.get(section, {})
-            pipfile_packages = set([
-                pep423_name(pkg_name) for pkg_name in pipfile_section.keys()
-            ])
+            pipfile_packages = set(
+                [pep423_name(pkg_name) for pkg_name in pipfile_section.keys()]
+            )
             to_remove = packages & pipfile_packages
             # The normal toml parser can't handle deleting packages with preceding newlines
             is_dev = section == "dev-packages"
@@ -1001,9 +930,7 @@ class Project(object):
         self.write_toml(p)
 
     def src_name_from_url(self, index_url):
-        name, _, tld_guess = six.moves.urllib.parse.urlsplit(index_url).netloc.rpartition(
-            "."
-        )
+        name, _, tld_guess = urllib.parse.urlsplit(index_url).netloc.rpartition(".")
         src_name = name.replace(".", "")
         try:
             self.get_source(name=src_name)
@@ -1011,6 +938,7 @@ class Project(object):
             name = src_name
         else:
             from random import randint
+
             name = "{0}-{1}".format(src_name, randint(1, 1000))
         return name
 
@@ -1018,20 +946,26 @@ class Project(object):
         """Adds a given index to the Pipfile."""
         # Read and append Pipfile.
         p = self.parsed_pipfile
+        source = None
         try:
-            self.get_source(url=index)
+            source = self.get_source(url=index)
         except SourceNotFound:
-            source = {"url": index, "verify_ssl": verify_ssl}
-        else:
-            return
+            try:
+                source = self.get_source(name=index)
+            except SourceNotFound:
+                pass
+        if source is not None:
+            return source["name"]
+        source = {"url": index, "verify_ssl": verify_ssl}
         source["name"] = self.src_name_from_url(index)
         # Add the package to the group.
         if "source" not in p:
-            p["source"] = [source]
+            p["source"] = [tomlkit.item(source)]
         else:
-            p["source"].append(source)
+            p["source"].append(tomlkit.item(source))
         # Write Pipfile.
         self.write_toml(p)
+        return source["name"]
 
     def recase_pipfile(self):
         if self.ensure_proper_casing():
@@ -1109,6 +1043,7 @@ class Project(object):
     @cached_property
     def finders(self):
         from .vendor.pythonfinder import Finder
+
         scripts_dirname = "Scripts" if os.name == "nt" else "bin"
         scripts_dir = os.path.join(self.virtualenv_location, scripts_dirname)
         finders = [
@@ -1130,3 +1065,29 @@ class Project(object):
             if as_path:
                 result = str(result.path)
         return result
+
+    def _which(self, command, location=None, allow_global=False):
+        if not allow_global and location is None:
+            if self.virtualenv_exists:
+                location = self.virtualenv_location
+            else:
+                location = os.environ.get("VIRTUAL_ENV", None)
+        if not (location and os.path.exists(location)) and not allow_global:
+            raise RuntimeError("location not created nor specified")
+
+        version_str = "python{}".format(".".join([str(v) for v in sys.version_info[:2]]))
+        is_python = command in ("python", os.path.basename(sys.executable), version_str)
+        if not allow_global:
+            if os.name == "nt":
+                p = find_windows_executable(os.path.join(location, "Scripts"), command)
+            else:
+                p = os.path.join(location, "bin", command)
+        else:
+            if is_python:
+                p = sys.executable
+        if not os.path.exists(p):
+            if is_python:
+                p = sys.executable or system_which("python")
+            else:
+                p = system_which(command)
+        return p
